from abc import ABC, abstractmethod
from typing import Optional

from prettytable import PrettyTable
from pydantic import AliasChoices, BaseModel, Field, computed_field, model_validator

from pycricinfo.models.output.common import SNAKE_CASE_REGEX, HeaderlessTableMixin
from pycricinfo.models.source.api.athelete import AthleteWithFirstAndLastName
from pycricinfo.models.source.api.linescores import BaseInningsDetails
from pycricinfo.models.source.api.team import TeamWithColorAndLogos

# ANSI escape codes for colors
RED = "\033[31m"
RESET = "\033[0m"


class LinescoreStatsLookupMixin(BaseModel, ABC):
    def add_linescore_stats_as_properties(data: dict, *args) -> dict:
        """
        Add items to the data dictionary so that extra keys can be deserialized into the Pydantic model.

        Find items by looking up the strings passed in as arguments, either matching to keys in the player's
        "general" statistics list for this innings, or to other parsed items in their batting or bowling innings.

        Parameters
        ----------
        data : dict
            The data to add keys to

        Returns
        -------
        dict
            The input data dictionary, with new keys added
        """
        linescore: BaseInningsDetails = data.get("linescore")
        if not linescore:
            return data

        for name in args:
            if not isinstance(name, str):
                raise TypeError("args to this function must be strings")
            name_split = str(name).split(".")
            stat_name = name_split[1] if len(name_split) > 1 else name_split[0]

            value = linescore.find(name)
            if value is not None:
                data[SNAKE_CASE_REGEX.sub("_", stat_name).lower()] = value
        return data


class PlayerInningsCommon(BaseModel, ABC):
    """
    Common base class shared between batting and bowling innings
    """

    order: int = Field(description="The order within the innings that this player batted or bowled")

    def colour_row(self, row_items: list[str], colour: str) -> list[str]:
        """
        Change the colour of a row in a PrettyTable.

        Parameters
        ----------
        row_items : list[str]
            Each cell in the row to be coloured.
        colour : str
            The ANSI escape code for the desired colour.

        Returns
        -------
        list[str]
            The row items with the specified colour applied.
        """
        return [f"{colour}{cell}{RESET}" for cell in row_items]

    @abstractmethod
    def add_to_table(self, table: PrettyTable, **kwargs): ...

    """ Abstract method which will be implemented in the Batting and Bowling innings classes """


class BattingInnings(PlayerInningsCommon):
    display_name: str
    dismissal_text: str
    captain: Optional[bool] = None
    keeper: Optional[bool] = None
    runs: int
    balls_faced: Optional[int] = None
    fours: Optional[int] = None
    sixes: Optional[int] = None
    minutes: Optional[int] = None
    not_out: bool = Field(validation_alias=AliasChoices("not_out", "notouts"))
    strike_rate: Optional[float | str] = Field(default=None, validation_alias=AliasChoices("strike_rate", "strikeRate"))

    @computed_field
    @property
    def player_display(self) -> str:
        """
        Get the batting scorecard display name of the player, including captain and keeper status.

        Returns
        -------
        str
            The player's display name, with captain and keeper indicators if applicable.
        """
        return f"{self.display_name}{' (c)' if self.captain else ''}{' \u271d' if self.keeper else ''}"

    def add_to_table(self, table: PrettyTable, **kwargs):
        """
        Add the batting innings details as in row in a PrettyTable, colouring the row red if the player is not out.

        Parameters
        ----------
        table : PrettyTable
            The PrettyTable instance to which the row will be added.
        """
        row_data = [
            self.player_display,
            self.dismissal_text,
            f"{self.runs}{'*' if self.not_out else ''}",
            self.balls_faced,
            self.fours,
            self.sixes,
        ]

        if kwargs.get("include_batting_minutes"):
            row_data.append(self.minutes)

        row_data.append(self.strike_rate)

        table.add_row(
            self.colour_row(
                row_data,
                RED if self.not_out else RESET,
            )
        )


class BowlingInnings(PlayerInningsCommon):
    display_name: str
    overs: float | int
    maidens: int
    runs: int = Field(validation_alias=AliasChoices("runs", "conceded"))
    wickets: int
    dots: Optional[int] = None
    no_balls: Optional[int] = Field(default=None, validation_alias=AliasChoices("no_balls", "noballs"))
    wides: Optional[int] = None
    economy_rate: Optional[float] = Field(validation_alias=AliasChoices("economy_rate", "economyRate"))
    fours_conceded: Optional[int] = Field(
        default=None, validation_alias=AliasChoices("fours_conceded", "foursConceded")
    )
    sixes_conceded: Optional[int] = Field(
        default=None, validation_alias=AliasChoices("sixes_conceded", "sixesConceded")
    )

    @computed_field
    @property
    def overs_display(self) -> float | int:
        """
        Round the overs to an integer if they are a whole number (to remove any '.0' on the end),
        otherwise return as a float.

        Returns
        -------
        float | int
            The overs bowled, rounded to an integer if applicable.
        """
        return int(self.overs) if self.overs % 1 == 0 else self.overs

    def add_to_table(self, table: PrettyTable, **kwargs):
        """
        Add the bowling innings details as a row in a PrettyTable.

        Parameters
        ----------
        table : PrettyTable
            The PrettyTable instance to which the row will be added.
        """
        row_data = [
            self.display_name,
            self.overs_display,
            self.maidens,
            self.runs,
            self.wickets,
            self.economy_rate,
            self.no_balls,
            self.wides,
            self.fours_conceded,
            self.sixes_conceded,
        ]

        if kwargs.get("include_bowling_dots"):
            row_data.append(self.dots)

        table.add_row(row_data)


class Innings(BaseModel, HeaderlessTableMixin):
    number: int
    batting_team_name: str
    batting_score: int
    declared: bool = Field(default=False)
    follow_on: bool = Field(default=False)
    wickets: int
    overs: Optional[float] = None
    batters: list[BattingInnings] = Field(default_factory=list)
    bowlers: list[BowlingInnings] = Field(default_factory=list)
    extras: Optional[int] = None
    byes: Optional[int] = None
    wides: Optional[int] = None
    leg_byes: Optional[int] = Field(default=None, validation_alias=AliasChoices("legbyes", "leg_byes"))
    no_balls: Optional[int] = Field(default=None, validation_alias=AliasChoices("noballs", "no_balls"))
    penalties: Optional[int] = None

    @computed_field
    @property
    def score_summary(self) -> str:
        """
        Get the score summary for the innings, including the score and wickets.

        Returns
        -------
        str
            The score summary in the format "<runs>/<wickets>" or "<runs> all out" as appropriate.
        """
        wickets_text = f" {'all out'}" if self.wickets == 10 else f"/{self.wickets}"
        overs_text = f" ({self.overs} overs)" if self.overs is not None else ""

        return (
            f"{self.batting_score}{wickets_text}"
            f"{'d' if self.declared else ''}{' (f/o)' if self.follow_on else ''}"
            f"{overs_text}"
        )

    @computed_field
    @property
    def extras_summary(self) -> str:
        """
        Get a summary of the extras for the innings, including byes, leg byes, wides, no balls,
        and (if present) penalties.

        Returns
        -------
        str
            A string summarizing the extras in the format:
            "Extras: <extras> (<byes>b <leg_byes>lb <wides>w <no_balls>nb <penalties>p)"
        """
        penalties_string = f" {self.penalties}p" if self.penalties and self.penalties != "0" else ""

        return (
            f"Extras: {self.extras or 0} "
            f"({self.byes or 0}b {self.leg_byes or 0}lb {self.wides or 0}w {self.no_balls or 0}nb{penalties_string})"
        )

    def show(self, **kwargs):
        """
        Print the innings details in PrettyTables. This will include the innings summary, followed by
        batting and bowling tables.
        """
        self.print_headerless_table(
            [
                (
                    f"Innings {self.number}: {self.batting_team_name} {self.score_summary}",
                    False,
                ),
                (
                    self.extras_summary,
                    False,
                ),
            ]
        )

        batting_field_names = ["", "Dismissal", "Runs", "Balls", "4s", "6s"]
        if kwargs.get("include_batting_minutes"):
            batting_field_names.append("Mins")
        batting_field_names.append("SR")

        self._print_player_innings_table(batting_field_names, self.batters, ["", "Dismissal"], **kwargs)

        bowling_field_names = ["", "Overs", "Maidens", "Runs", "Wickets", "Economy", "No Balls", "Wides", "4s", "6s"]
        if kwargs.get("include_bowling_dots"):
            bowling_field_names.append("Dots")

        self._print_player_innings_table(bowling_field_names, self.bowlers, [""], **kwargs)

    def _print_player_innings_table(
        self,
        field_names: list[str],
        items: list[PlayerInningsCommon],
        field_names_to_left_align: list[str] = None,
        **kwargs,
    ):
        """
        Print a PrettyTable with the specified field names and items, representing either a Bowling or Batting innings.

        Parameters
        ----------
        field_names : list[str]
            The names of the fields to be displayed in the table.
        items : list[PlayerInningsCommon]
            The list of player innings items to be added to the table.
        field_names_to_left_align : list[str], optional
            Which fields in the table to align to the left (rather than the usual centre), by default None
        """
        table = PrettyTable()
        table.field_names = field_names
        for name in field_names_to_left_align or []:
            table.align[name] = "l"

        for player in sorted(items, key=lambda b: b.order):
            player.add_to_table(table, **kwargs)
        print(table)


class CricinfoBattingInnings(BattingInnings, LinescoreStatsLookupMixin):
    player: AthleteWithFirstAndLastName  # Could be full Athlete

    @model_validator(mode="before")
    @classmethod
    def create_batting_attributes(cls, data: dict) -> dict:
        """
        Run before Pydantic validation to create the required fields in the data dictionary.

        Find the batting statistics in the linescore and add them as properties to the data dictionary.

        Parameters
        ----------
        data : dict
            The input data being validated into this model. It should contain a "linescore" key with a
            BaseInningsDetails object.

        Returns
        -------
        dict
            The transformed data dictionary with the required fields for a CricinfoBattingInnings, which Pydantic can
            now validate.
        """
        data = cls.add_linescore_stats_as_properties(
            data,
            "batting.dismissal_text",
            "runs",
            "ballsFaced",
            "notouts",
            "batting.order",
            "fours",
            "sixes",
            "minutes",
            "strikeRate",
        )
        return data


class CricinfoBowlingInnings(BowlingInnings, LinescoreStatsLookupMixin):
    player: AthleteWithFirstAndLastName  # Could be full Athlete

    @model_validator(mode="before")
    @classmethod
    def create_bowling_attributes(cls, data: dict) -> dict:
        """
        Run before Pydantic validation to create the required fields in the data dictionary.

        Find the bowling statistics in the linescore and add them as properties to the data dictionary.

        Parameters
        ----------
        data : dict
            The input data being validated into this model. It should contain a "linescore" key with a
            BaseInningsDetails object.

        Returns
        -------
        dict
            The transformed data dictionary with the required fields for a CricinfoBowlingInnings, which Pydantic can
            now validate.
        """
        return cls.add_linescore_stats_as_properties(
            data,
            "overs",
            "maidens",
            "conceded",
            "wickets",
            "bowling.order",
            "dots",
            "noballs",
            "wides",
            "foursConceded",
            "sixesConceded",
            "economyRate",
        )


class CricinfoInnings(Innings, LinescoreStatsLookupMixin):
    team: TeamWithColorAndLogos

    @model_validator(mode="before")
    @classmethod
    def create_innings_attributes(cls, data: dict) -> dict:
        """
        Run before Pydantic validation to create the required fields in the data dictionary.

        Find the innings statistics in the linescore and add them as properties to the data dictionary.

        Parameters
        ----------
        data : dict
            The input data being validated into this model. It should contain a "linescore" key with a
            BaseInningsDetails object.

        Returns
        -------
        dict
            The transformed data dictionary with the required fields for a CricinfoInnings, which Pydantic can
            now validate.
        """
        return cls.add_linescore_stats_as_properties(
            data,
            "bpo",
            "byes",
            "extras",
            "legbyes",
            "noballs",
            "penalties",
            "runRate",
            "wides",
        )



from typing import Annotated

from fastapi import APIRouter, Depends, Path, status

from pycricinfo.api.utils import PageAndInningsQueryParameters
from pycricinfo.call_cricinfo_api import get_play_by_play
from pycricinfo.models.source.api.commentary import Commentary

router = APIRouter(prefix="", tags=["play_by_play"])


@router.get(
    "/match/{match_id}/play_by_play",
    responses={status.HTTP_200_OK: {"description": "The match summary"}},
    summary="Get a page of ball-by-ball data",
)
def match_play_by_play_api(
    match_id: Annotated[int, Path(description="The Match ID")], pi: PageAndInningsQueryParameters = Depends()
) -> Commentary:
    return get_play_by_play(match_id, pi.innings, pi.page)



from typing import Optional

from pydantic import BaseModel

from pycricinfo.models.source.api.athelete import AthleteWithFirstAndLastName
from pycricinfo.models.source.api.common import CCBaseModel


class DismissalFielder(CCBaseModel):
    athlete: AthleteWithFirstAndLastName
    display_order: int
    is_keeper: int
    is_substitute: int


class DismissalDetailsInnings(BaseModel):
    wickets: int
    runs: int


class DismissalDetailsOver(BaseModel):
    overs: float


class DismissalDetails(CCBaseModel):
    id: str
    text: str
    short_text: str
    innings: DismissalDetailsInnings
    over: DismissalDetailsOver


class Dismissal(CCBaseModel):
    bowler: Optional[AthleteWithFirstAndLastName] = None
    details: Optional[DismissalDetails] = None
    dismissal_card: str
    fielders: list[DismissalFielder]
    short_text: str



from typing import Optional

from pydantic import BaseModel, Field


class MatchSeries(BaseModel):
    series_id: int
    data_series_id: int
    title: str
    link: str
    summary_url: str


class MatchTypeWithSeries(BaseModel):
    name: str
    series: Optional[list[MatchSeries]] = Field(default_factory=list)


class MatchResult(BaseModel):
    id: int
    description: str
    innings_1_info: Optional[str] = None
    innings_2_info: Optional[str] = None
    status: Optional[str] = None



from argparse import ArgumentParser, Namespace

from pydantic import ValidationError

from pycricinfo.call_cricinfo_api import get_match, get_play_by_play
from pycricinfo.models.output.scorecard import CricinfoScorecard
from pycricinfo.models.source.api.commentary import APIResponseCommentary, Commentary
from pycricinfo.models.source.api.match import Match
from pycricinfo.utils import load_file_and_validate_to_model


def print_scorecard(file_path: str = None, match_id: int = None, series_id: int = None):
    """
    Prints the scorecard of a match, either by passing a file path or loading from command line arguments

    Parameters
    ----------
    file_path : str, optional
        The path to a JSON file containing match data. If not provided, it will be taken from command line arguments.
    """
    args = parse_scorecard_args()

    if file_path or args.file:
        _print_scorecard_from_file(file_path or args.file)
    elif match_id or args.match_id:
        _print_scorecard_from_match_id(series_id or args.series_id, match_id or args.match_id)
    else:
        print("Please provide either a file path or match & series IDs")


def _print_scorecard_from_file(file_path: str):
    model = load_file_and_validate_to_model(file_path, Match)
    _print_scorecard_from_match(model)


def _print_scorecard_from_match_id(series_id: int, match_id: int):
    model = get_match(series_id, match_id)
    _print_scorecard_from_match(model)


def _print_scorecard_from_match(match: Match):
    try:
        sc = CricinfoScorecard(match=match)
        sc.show(include_batting_minutes=False, include_bowling_dots=False)
    except ValidationError as validation_error:
        print(validation_error.errors())
        raise


def parse_scorecard_args() -> Namespace:
    """
    Parse command line arguments

    Returns
    -------
    argparse.Namespace
        The parsed arguments
    """
    parser = ArgumentParser()
    parser.add_argument("--file", help="Path to a JSON file to parse and print from")
    parser.add_argument("--series_id", help="ID of the series of the match to fetch from the API")
    parser.add_argument("--match_id", help="ID of the match to fetch from the API")
    args = parser.parse_args()
    return args


def print_ball_by_ball(file_path: str = None, match_id: int = None, innings: int = 1, page: int = 1):
    """
    Prints a page of ball by ball commentary of a match, either by passing a file path or loading from command line
    arguments

    Parameters
    ----------
    file_path : str, optional
        The path to a JSON file containing match data. If not provided, it will be taken from command line arguments.
    match_id : int, optional
        The ID of the match to fetch from the API. If not provided, it will be taken from command line arguments.
    innings : int, optional
        The innings of the match to fetch commentary for. Defaults to 1.
    page : int, optional
        The page of commentary to fetch. Defaults to 1.
    """
    args = parse_ball_by_ball_args()

    if file_path or args.file:
        _print_ball_by_ball_from_file(file_path or args.file)
    elif match_id or args.match_id:
        match_id = match_id or args.match_id
        innings = innings or args.innings
        page = page or args.page
        _print_ball_by_ball_from_match_id(match_id, innings, page)
    else:
        print("Please provide either a file path or match ID and innings/page parameters.")


def _print_ball_by_ball_from_file(file_path: str):
    model = load_file_and_validate_to_model(file_path, APIResponseCommentary)
    _print_ball_by_ball_from_commentary_model(model.commentary)


def _print_ball_by_ball_from_match_id(match_id: int, innings: int, page: int):
    model = get_play_by_play(match_id, innings, page)
    _print_ball_by_ball_from_commentary_model(model)


def _print_ball_by_ball_from_commentary_model(model: Commentary):
    for item in model.deliveries:
        print(f"{item.over.overs}: {item.short_text} - {item.current_innings_score.score}")


def parse_ball_by_ball_args() -> Namespace:
    """
    Parse command line arguments

    Returns
    -------
    argparse.Namespace
        The parsed arguments
    """
    parser = ArgumentParser()
    parser.add_argument("--file", help="Path to a JSON file to parse and print from")
    parser.add_argument("--match_id", help="ID of the match to fetch from the API")
    parser.add_argument("--innings", help="The innings of the game to get data from", type=int, default=1)
    parser.add_argument("--page", help="The page of commentary to return from that innings", type=int, default=1)
    args = parser.parse_args()
    return args



import re

from prettytable import PrettyTable

SNAKE_CASE_REGEX = re.compile(r"(?<!^)(?=[A-Z])")


class HeaderlessTableMixin:
    def print_headerless_table(self, rows: list[tuple[str, bool]]):
        """
        Print a PrettyTable with no header, with rows matching the supplied data

        Parameters
        ----------
        rows : list[tuple[str, bool]]
            The rows of data to print. Each tuple contains the data to print,
            and whether to include a divider after the row (True) or not (False).
        """
        table = PrettyTable()
        table.header = False
        for row in rows:
            table.add_row([row[0]], divider=row[1])
        print(table)



from pycricinfo.api_helper import get_and_parse, get_request
from pycricinfo.config import BaseRoute, get_settings
from pycricinfo.models.output.scorecard import CricinfoScorecard
from pycricinfo.models.source import APIResponseCommentary, Commentary, Match, MatchBasic, Player, TeamFull


def get_player(player_id: int) -> Player:
    """
    Get a player by their ID.

    Parameters
    ----------
    player_id : int
        The ID of the player to retrieve.

    Returns
    -------
    Player
        A parsed Pydantic model representing the player.
    """
    return get_and_parse(get_settings().routes.player, Player, params={"player_id": player_id})


def get_team(team_id: int) -> TeamFull:
    """
    Get a team by its ID.

    Parameters
    ----------
    team_id : int
        The ID of the team to retrieve.

    Returns
    -------
    TeamFull
        A parsed Pydantic model representing the team.
    """
    # TODO: Format "classes" field onto what match type they represent
    return get_and_parse(get_settings().routes.team, TeamFull, params={"team_id": team_id})


def get_match_basic(match_id: int) -> MatchBasic:
    """
    Get basic match information by match ID.

    Parameters
    ----------
    match_id : int
        The ID of the match to retrieve.

    Returns
    -------
    MatchBasic
        A parsed Pydantic model representing the basic match information.
    """
    return get_and_parse(get_settings().routes.match_basic, MatchBasic, params={"match_id": match_id})


def get_match(series_id: int, match_id: int) -> Match:
    """
    Get detailed match information by match ID.

    Parameters
    ----------
    series_id : int
        The ID of the series to which the match belongs.

    match_id : int
        The ID of the match to retrieve.

    Returns
    -------
    Match
        A parsed Pydantic model representing the match details.
    """
    return get_and_parse(
        get_settings().routes.match_summary,
        Match,
        params={"series_id": series_id, "match_id": match_id},
        base_route=BaseRoute.site,
    )


def get_match_raw(series_id: int, match_id: int) -> dict:
    """
    Get raw match data by match ID.

    Parameters
    ----------
    series_id : int
        The ID of the series to which the match belongs.

    match_id : int
        The ID of the match to retrieve.

    Returns
    -------
    dict
        The raw match data as a dictionary.
    """
    return get_request(
        get_settings().routes.match_summary,
        params={"series_id": series_id, "match_id": match_id},
        base_route=BaseRoute.site,
        response_output_sub_folder="matches",
    )


def get_scorecard(series_id: int, match_id: int) -> CricinfoScorecard:
    """
    Get a match and generate and return a scorecard for it.

    Parameters
    ----------
    match_id : int
        The ID of the match for which to generate the scorecard.

    Returns
    -------
    Scorecard
        A scorecard object containing match details and scores.
    """
    match = get_match(series_id, match_id)
    return CricinfoScorecard(match=match)


def get_play_by_play(match_id: int, innings: int = 1, page: int = 1) -> Commentary:
    """
    Get a page of ball-by-ball data for a match, processed into a list of CommentaryItems.

    Parameters
    ----------
    match_id : int
        The ID of the match for which to retrieve ball-by-ball commentary.
    page : int, optional
        The page of commentary to return, by default 1
    innings : int, optional
        Which innings to retrieve commentary for, by default 1

    Returns
    -------
    list[CommentaryItem]
        A list of parsed Pydantic models representing the ball-by-ball commentary.
    """
    response = get_and_parse(
        get_settings().routes.play_by_play_page,
        APIResponseCommentary,
        {"match_id": match_id, "page": page, "innings": innings},
        True,
        BaseRoute.site,
    )
    return response.commentary if response and response.commentary else []


def get_play_by_play_raw(match_id: int, page: int = 1, innings: int = 1) -> dict:
    """
    Get a page of ball-by-ball data for a match.

    Parameters
    ----------
    match_id : int
        The ID of the match for which to retrieve ball-by-ball commentary.
    page : int, optional
        The page of commentary to return, by default 1
    innings : int, optional
        Which innings to retrieve commentary for, by default 1

    Returns
    -------
    dict
        The raw play-by-play data as a dictionary.
    """
    return get_request(
        get_settings().routes.play_by_play_page,
        {"match_id": match_id, "page": page, "innings": innings},
        base_route=BaseRoute.site,
        response_output_sub_folder="play_by_play",
    )



import os

import pytest

from pycricinfo.models.source import Match, MatchBasic
from pycricinfo.utils import load_file_and_validate_to_model


def get_test_file_path(file_name: str) -> str:
    """Get the full path to a test file."""
    return os.path.join(os.path.dirname(__file__), "test_files", file_name)


@pytest.mark.parametrize("file_name", ["match/1426555.json", "match/1243385.json", "match/1381212.json"])
def test_load_match(file_name):
    """Test loading JSON files into Match model without exceptions."""
    test_file_path = get_test_file_path(file_name)

    result = load_file_and_validate_to_model(test_file_path, Match)

    assert result is not None
    assert isinstance(result, Match)


def test_load_match_basic():
    """Test loading JSON files into MatchBasic model without exceptions."""
    test_file_path = get_test_file_path("match_basic/1225249_basic.json")

    result = load_file_and_validate_to_model(test_file_path, MatchBasic)

    assert result is not None
    assert isinstance(result, MatchBasic)



from typing import Optional

from pydantic import AliasChoices, BaseModel, Field, model_validator

from pycricinfo.models.source.api.common import CCBaseModel


class BowlingDetailsToHand(CCBaseModel):
    balls: int
    wickets: int
    economy_rate: float
    conceded: int


class PitchMapElement(BaseModel):
    runs: int
    wickets: int
    balls: int


class PitchMapLength(BaseModel):
    wide_outside_off: PitchMapElement
    outside_off: PitchMapElement
    straight: PitchMapElement
    outside_leg: PitchMapElement
    wide_outside_leg: PitchMapElement


class PitchMap(BaseModel):
    full_toss: PitchMapLength
    yorker: PitchMapLength
    full: PitchMapLength
    good: PitchMapLength
    short_of_good: PitchMapLength
    short: PitchMapLength


class BowlingDetails(CCBaseModel):
    active: bool
    active_name: str
    order: int
    overall_lhb: Optional[BowlingDetailsToHand] = None
    overall_rhb: Optional[BowlingDetailsToHand] = None
    pitch_map_lhb_raw: Optional[list[list[list[int]]]] = Field(
        default=None, validation_alias=AliasChoices("pitch_map_lhb_raw", "pitchMapLhb")
    )
    pitch_map_rhb_raw: Optional[list[list[list[int]]]] = Field(
        default=None, validation_alias=AliasChoices("pitch_map_rhb_raw", "pitchMapRhb")
    )
    pitch_map_right: Optional[PitchMap] = None
    pitch_map_left: Optional[PitchMap] = None

    @model_validator(mode="before")
    @classmethod
    def generate_structured_pitch_maps(cls, data: dict):
        pitch_map_length_fields = PitchMapLength.model_fields.keys()
        pitch_map_fields = PitchMap.model_fields.keys()

        pitchMapRhb = data.get("pitchMapRhb", None)
        if pitchMapRhb:
            data["pitchMapRight"] = cls._generate_pitch_map(pitch_map_fields, pitch_map_length_fields, pitchMapRhb)

        pitchMapLhb = data.get("pitchMapLhb", None)
        if pitchMapLhb:
            data["pitchMapLeft"] = cls._generate_pitch_map(pitch_map_fields, pitch_map_length_fields, pitchMapLhb)
        return data

    def _generate_pitch_map(
        pitch_map_fields: list[str], pitch_map_length_fields: list[str], raw_pitch_map_data: list[list[list[int]]]
    ):
        lengths = []
        for length in raw_pitch_map_data:
            lines = []
            for line in length:
                element = PitchMapElement(runs=line[0], wickets=line[1], balls=line[2])
                lines.append(element)

            line_data = dict(zip(pitch_map_length_fields, lines))
            length_map = PitchMapLength(**line_data)
            lengths.append(length_map)

        length_data = dict(zip(pitch_map_fields, lengths))
        return PitchMap(**length_data)



from abc import ABC
from typing import Optional

from pydantic import AliasChoices, Field

from pycricinfo.models.source.api.common import CCBaseModel, Event, RefMixin


class TeamCommon(CCBaseModel, ABC):
    id: str
    abbreviation: str
    display_name: str


class TeamWithName(TeamCommon):
    name: str = Field(description="The full name of the Team")


class TeamWithColorAndLogos(TeamCommon):
    color: str = Field(description="The hex colour for the Team", examples=["#790d1a"])
    logos: list[RefMixin]


class TeamFull(TeamWithName):
    color: str
    nickname: str
    short_display_name: str
    is_national: bool
    is_active: bool
    classes: list[int] = Field(description="The classes of match that this Team plays in")
    current_match: Event = Field(default=None, validation_alias=AliasChoices("event"))
    current_players_link: RefMixin = Field(default=None, validation_alias=AliasChoices("athletes"))


class TeamWicketDetails(CCBaseModel):
    text: str
    short_text: str


class TeamWicket(CCBaseModel):
    details: TeamWicketDetails
    balls_faced: int
    dismissal_card: str
    fours: int
    fow: str
    minutes: Optional[int | str] = None  # TODO: Can be empty string - parse to null in that case
    number: int
    over: float
    runs: int
    short_text: str
    sixes: int
    strike_rate: float | str


class TeamOver(CCBaseModel):
    number: int
    runs: int
    wicket: list[TeamWicket]



from abc import ABC
from typing import Optional

from pydantic import BaseModel, computed_field

from pycricinfo.models.source.api.batting import BattingDetails
from pycricinfo.models.source.api.bowling import BowlingDetails
from pycricinfo.models.source.api.common import CCBaseModel
from pycricinfo.models.source.api.team import TeamOver


class BasicStatistic(CCBaseModel):
    name: str
    display_value: Optional[str] = None
    value: int | str | float


class StatsCategory(BaseModel):
    name: str
    stats: list[BasicStatistic]

    batting_stat_names: list[str] = [
        "ballsFaced",
        "batted",
        "battingPosition",
        "dismissal",
        "dismissalCard",
        "ducks",
        "fielderKeeper",
        "fielderSub",
        "fiftyPlus",
        "fours",
        "hundreds",
        "innings",
        "inningsNumber",
        "minutes",
        "notouts",
        "outs",
        "retiredDescription",
        "runs",
        "sixes",
        "strikeRate",
    ]

    bowling_stat_names: list[str] = [
        "balls",
        "bowled",
        "bowlingPosition",
        "bpo",
        "caught",
        "caughtFielder",
        "caughtKeeper",
        "conceded",
        "dismissals",
        "dots",
        "economyRate",
        "fielded",
        "fiveWickets",
        "fourPlusWickets",
        "foursConceded",
        "illegalOverLimit",
        "inningsBowled",
        "inningsFielded",
        "inningsNumber",
        "maidens",
        "noballs",
        "overs",
        "sixesConceded",
        "stumped",
        "tenWickets",
        "wickets",
        "wides",
    ]

    team_stat_names: list[str] = [
        "ballLimit",
        "balls",
        "bpo",
        "byes",
        "extras",
        "lead",
        "legbyes",
        "liveCurrent",
        "minutes",
        "miscounted",
        "noballs",
        "oldPenaltyOrBonus",
        "overLimit",
        "overLimitRunRate",
        "overSplitLimit",
        "overs",
        "oversDocked",
        "penalties",
        "penaltiesFieldEnd",
        "penaltiesFieldStart",
        "runRate",
        "runs",
        "target",
        "targetachievedMinutes",
        "targetachievedOvers",
        "targetachievedRuns",
        "targetachievedWickets",
        "wickets",
        "wides",
    ]

    def get_stat(self, name: str) -> int | str | float:
        return next((s.value for s in self.stats if s.name == name), None)


class StatisticsCategory(ABC, BaseModel):
    id: str
    name: str
    abbreviation: str
    categories: list[StatsCategory]

    @computed_field
    @property
    def general_category(self) -> StatsCategory:
        first = next(iter(self.categories), None)
        return first

    def find(self, name: str) -> int | str | float:
        return self.general_category and self.general_category.get_stat(name)


class PlayerStatisticsCategory(StatisticsCategory):
    batting: Optional[BattingDetails] = None
    bowling: Optional[BowlingDetails] = None

    def find(self, name: str) -> int | str | float:
        split: list[str] = name.split(".")
        if len(split) == 1:
            return self.general_category and self.general_category.get_stat(name)

        if split[0] == ("batting"):
            return self.batting and getattr(self.batting, split[1])

        if split[0] == ("bowling"):
            return self.bowling and getattr(self.bowling, split[1])


class TeamStatisticsCategory(StatisticsCategory):
    overs: list[list[TeamOver]]



import json
import logging
import re
import uuid
from datetime import UTC, datetime
from pathlib import Path
from typing import Optional, Type, TypeVar
from urllib.parse import urljoin, urlparse

import requests
from pydantic import BaseModel, ValidationError

from pycricinfo.config import BaseRoute, get_settings
from pycricinfo.exceptions import CricinfoAPIException
from pycricinfo.utils import replace_empty_objects_with_null

logger = logging.getLogger("cricinfo")
T = TypeVar("T", bound=BaseModel)


def get_and_parse(
    route: str,
    type_to_parse: Type[T],
    params: dict = None,
    null_out_empty_dicts: bool = False,
    base_route: BaseRoute = BaseRoute.core,
) -> T:
    """
    Make a GET request to the API and parse the response to the supplied type

    Parameters
    ----------
    route : str
        The route template to call
    type_to_parse : Type[T]
        The source model type to attempt to parse the response into
    params : dict, optional
        Any parameters to fill in into the route, in a dictionary of key-value pairs, by default None
    null_out_empty_dicts : bool, optional
        Whether to replace any dictionaries that contain only None values with None, by default False
    base_route: BaseRoute, optional
        The base route to use for the API call, by default BaseRoute.core
    Returns
    -------
    T
        The response data parsed into the supplied model
    """
    api_response = get_request(route, params, base_route)

    if null_out_empty_dicts:
        api_response = replace_empty_objects_with_null(api_response)

    try:
        return type_to_parse.model_validate(api_response)
    except ValidationError as ex:
        logger.error(ex)
        raise


def get_request(
    route: str,
    params: Optional[dict[str, str]] = None,
    base_route: BaseRoute = BaseRoute.core,
    response_output_sub_folder: str = None,
) -> dict | str:
    """
    Make a GET request to the Football Stats API

    Parameters
    ----------
    route : str
        The route template to call
    params : dict[str, str], optional
        Any parameters to fill in into the route, in a dictionary of key-value pairs, by default None
    base_route: BaseRoute, optional
        The base route to use for the API call, by default BaseRoute.core
    Returns
    -------
    dict
        The JSON content of the API response
    """
    request_id = str(uuid.uuid4())
    response_logging_extras = {
        "cricket_stats.request_id": request_id,
        "cricket_stats.request_route_template": route,
    }

    if params:
        route = _format_route(route, params)

    if base_route == BaseRoute.core:
        base = get_settings().core_base_route_v2
    elif base_route == BaseRoute.site:
        base = get_settings().site_base_route_v2
    else:
        base = get_settings().pages_base_route
    full_route = f"{base}{route}"

    session = requests.Session()
    session.headers["User-Agent"] = get_settings().page_headers.user_agent
    session.headers["Referer"] = urljoin(route, urlparse(route).path)
    session.headers["Accept"] = get_settings().page_headers.accept

    logger.debug(f"Querying: {full_route}", extra={"cricket_stats.request_id": request_id})
    response = session.get(full_route)

    response_logging_extras["cricket_stats.response_code"] = response.status_code
    if base_route == BaseRoute.page:
        result: bytes = response.content
        logger.debug(json.dumps(f"Page fetched from: {full_route}"), extra=response_logging_extras)
        output = str(result)
        output_for_file = re.sub(r"^b\'|\'$", "", output)
        response_output_file_extension = "html"
    else:
        output = result = response.json()
        logger.debug(json.dumps(result, indent=4), extra=response_logging_extras)
        output_for_file = json.dumps(result, indent=4)
        response_output_file_extension = "json"

    _output_response_to_file(output_for_file, route, response_output_sub_folder, response_output_file_extension)

    if response.status_code != 200:
        logger.error(
            f"Status Code '{response.status_code}' returned for '{full_route}'",
            extra=response_logging_extras,
        )
        raise CricinfoAPIException(status_code=response.status_code, route=full_route, content=output)

    return output


def _format_route(route: str, params: dict[str, str] = {}) -> str:
    """
    Format the route with the provided parameters

    Parameters
    ----------
    route : str
        The route template to format
    params : dict[str, str]
        The parameters to fill in into the route

    Returns
    -------
    str
        The formatted route
    """
    for key, value in params.items():
        if not key.startswith("{"):
            key = "{" + key
        if not key.endswith("}"):
            key = key + "}"
        route = route.replace(key, str(value))
    return route


def _output_response_to_file(response: str, route: str, sub_folder: str, file_extension: str) -> None:
    """
    Output the content of the response to a file

    Parameters
    ----------
    response : str
        The API/page response
    route : str
        The route that was called
    """
    folder = Path(get_settings().api_response_output_folder)
    today = datetime.today().strftime("%Y%m%d")
    file_name = f"{datetime.now(UTC).time().strftime('%H%M%S')}_{route.replace('/', '_')}.{file_extension}"

    if sub_folder:
        folder = folder / sub_folder
    folder = folder / today
    folder.mkdir(parents=True, exist_ok=True)

    file_path = (folder / file_name).resolve()
    with open(file_path, "w") as file:
        file.write(response)



from typing import Optional

from pydantic import BaseModel, model_validator

from pycricinfo.models.output.common import HeaderlessTableMixin
from pycricinfo.models.source.api.commentary import APIResponseCommentary


class BallByBallPage(BaseModel, HeaderlessTableMixin):
    match_title: Optional[str]
    details: Optional[str]
    deliveries: list

    @model_validator(mode="before")
    @classmethod
    def create(cls, data: dict):
        page: APIResponseCommentary = data["page"]

        for item in page.commentary.items:
            pass



from pycricinfo.call_cricinfo_api import (
    get_match,
    get_match_basic,
    get_play_by_play,
    get_player,
    get_scorecard,
    get_team,
)

__all__ = [
    "get_match",
    "get_match_basic",
    "get_play_by_play",
    "get_player",
    "get_scorecard",
    "get_team",
]



from typing import Annotated

import requests
from fastapi import APIRouter, Path, status

from pycricinfo.call_cricinfo_api import get_match, get_match_basic
from pycricinfo.models.source.api.match import Match
from pycricinfo.models.source.api.match_basic import MatchBasic

router = APIRouter(prefix="/match", tags=["match"])


@router.get(
    "/{match_id}",
    responses={status.HTTP_200_OK: {"description": "The basic match data"}},
    summary="Get basic match data from the '/events' API",
)
def match_basic(match_id: int = Path(description="The Match ID")) -> MatchBasic:
    return get_match_basic(match_id)


@router.get(
    "/{match_id}/team/{team_id}",
    responses={status.HTTP_200_OK: {"description": "The basic match data"}},
    summary="Get a match's Team",
)
def get_match_team(match_id: int = Path(description="The Match ID"), team_id: int = Path(description="The Team ID")):
    response = requests.get(
        f"http://core.espnuk.org/v2/sports/cricket/leagues/0/events/{match_id}/competitions/{match_id}/competitors/{team_id}"
    ).json()
    return response


@router.get(
    "/summary/{series_id}/{match_id}",
    responses={status.HTTP_200_OK: {"description": "The match summary"}},
    summary="Get a match summary",
)
def match_api(
    series_id: Annotated[int, Path(description="The Series ID")],
    match_id: Annotated[int, Path(description="The Match ID")],
) -> Match:
    return get_match(series_id, match_id)



import importlib.util
import json
from importlib.metadata import metadata
from typing import Any, Type, TypeVar

from pydantic import BaseModel, ValidationError

T = TypeVar("T", bound=BaseModel)


def replace_empty_objects_with_null(data: Any) -> Any:
    """
    Recursively replace any dictionaries or lists that contain only None values with None.
    Any data that is not a dictionary or list will be returned as is.

    Parameters
    ----------
    data : Any
        The data to check for empty dictionaries or lists

    Returns
    -------
    Any
        The data with empty dictionaries or lists replaced with None
    """
    if isinstance(data, dict):
        if all(value is None for value in data.values()):
            return None
        return {key: replace_empty_objects_with_null(value) for key, value in data.items()}
    elif isinstance(data, list):
        return [replace_empty_objects_with_null(item) for item in data]
    else:
        return data


def load_file_and_validate_to_model(file_path: str, type_to_parse: Type[T]) -> T:
    """
    Load a JSON file from disk and validate its contents against a Pydantic model.

    Parameters
    ----------
    file_path : str
        The path to the JSON file to load
    type_to_parse : Type[T]
        The Pydantic model to validate against

    Returns
    -------
    T
        An instance of the Pydantic model with the loaded data
    """
    with open(file_path, "r") as content:
        json_data = json.load(content)
        return load_dict_to_model(json_data, type_to_parse)


def load_dict_to_model(json_data: dict, type_to_parse: Type[T]) -> T:
    """
    Validate a dictionary against a Pydantic model.

    Parameters
    ----------
    json_data : dict
        The dictionary to validate
    type_to_parse : Type[T]
        The Pydantic model to validate against

    Returns
    -------
    T
        An instance of the Pydantic model with the loaded data
    """
    modified_data = replace_empty_objects_with_null(json_data)
    try:
        model = type_to_parse.model_validate(modified_data)
    except ValidationError as ex:
        print(ex)
        exit(1)

    return model


def get_field_from_pyproject(field_name: str) -> str:
    """
    Get a specific field from package metadata.

    Parameters
    ----------
    field_name : str
        The name of the field to retrieve (e.g., "version", "description")

    Returns
    -------
    str
        The value of the specified field from package metadata
    """
    pyproject_to_importlib_field_mappings = {
        "version": "Version",
        "description": "Summary",
    }

    if importlib.util.find_spec("importlib.metadata") is not None:
        try:
            pkg_metadata = metadata("pycricinfo")
            mapped_field = pyproject_to_importlib_field_mappings.get(field_name, field_name.capitalize())
            value = pkg_metadata.get(mapped_field, "")
            if value:
                return value
        except (ImportError, KeyError):
            pass

    return ""



import html
from typing import Optional

from pydantic import BaseModel, Field, model_validator

from pycricinfo.models.output.common import HeaderlessTableMixin
from pycricinfo.models.output.innings import CricinfoBattingInnings, CricinfoBowlingInnings, CricinfoInnings, Innings
from pycricinfo.models.source.api.match import Match
from pycricinfo.models.source.api.roster import MatchPlayer, TeamLineup


class Scorecard(BaseModel, HeaderlessTableMixin):
    title: Optional[str] = Field(
        description="The title for the scorecard, usually match details, including teams"
        "and dates. e.g.) 3rd Test, West Indies tour of England at Birmingham, Jul 26-28 2024 "
    )
    summary: Optional[str] = Field(description="A summary of the result of the match, e.g.) 'India won by 5 wickets'")
    innings: list[Innings]

    def show(self, **kwargs):
        """
        Print the scorecard as a PrettyTable, including the title and summary.

        Parameters
        ----------
        kwargs
            Supports parameters to customise the columns of the scorecard table:
            - include_batting_minutes : bool
            - include_bowling_dots : bool
        """
        self.print_headerless_table([(self.title, True), (self.summary, False)])

        for innings in self.innings:
            innings.show(**kwargs)


class CricinfoScorecard(Scorecard):
    @model_validator(mode="before")
    @classmethod
    def create(cls, data: dict) -> dict:
        """
        Run before Pydantic validation to create the required fields in the data dictionary.

        Transforms input data into a Scorecard, extracting details from a Match object.

        Parameters
        ----------
        data : dict
            The input data being validated into this model. It should contain a "match" key with a Match object.

        Raises
        ------
        ValueError
            If the "match" key is not present in the input data.

        Returns
        -------
        dict
            The transformed data dictionary with the required fields for a CricinfoScorecard, which Pydantic can
            now validate.
        """
        match: Match = data["match"]
        if not match:
            raise ValueError("Match data is required to create a CricinfoScorecard.")

        data["title"] = html.unescape(match.header.title)
        data["summary"] = html.unescape(match.summary)

        innings = []
        for i in range(1, 3 if match.header.competition.limited_overs else 5):
            response = match.header.get_batting_innings_by_number(i)
            if not response:
                continue

            team, linescore = response
            innings.append(
                CricinfoInnings(
                    number=i,
                    team=team,
                    batting_team_name=team.display_name,
                    batting_score=linescore.runs,
                    wickets=linescore.wickets,
                    overs=linescore.overs,
                    linescore=linescore,
                    declared=linescore.declared,
                    follow_on=linescore.follow_on,
                )
            )
        for roster in match.rosters:
            cls._enrich_innings_with_lineups(innings, roster)

        data["innings"] = innings
        return data

    @classmethod
    def _enrich_innings_with_lineups(cls, innings: list[CricinfoInnings], lineup: TeamLineup):
        """
        Enrich the innings with player data from the team lineup. The innings will be updated in
        place, adding batters and bowlers based on the players in the lineup.

        Parameters
        ----------
        innings : list[CricinfoInnings]
            All innings in the match, which will be enriched with player data.
        lineup : TeamLineup
            The team lineup containing player data to enrich the innings with.
        """
        for player in lineup.players:
            cls._enrich_innings_for_player(innings, player)

    @classmethod
    def _enrich_innings_for_player(cls, innings: list[CricinfoInnings], player: MatchPlayer):
        """
        Enrich the innings with data for a specific player. The innings will be updated in
        place, adding batters or bowlers records based on this player's data.

        Parameters
        ----------
        innings : list[CricinfoInnings]
            All innings in the match, which will be enriched with player data.
        player : MatchPlayer
            The player whose data will be used to enrich the innings.
        """
        for linescore in player.linescores:
            if bool(linescore.batted) and bool(int(linescore.batted)):
                bat = CricinfoBattingInnings(
                    player=player.athlete,
                    display_name=player.athlete.display_name,
                    captain=player.captain,
                    keeper=player.keeper,
                    linescore=linescore,
                )
                innings[linescore.period - 1].batters.append(bat)
            elif bool(linescore.bowled) and bool(int(linescore.bowled)):
                bowl = CricinfoBowlingInnings(
                    player=player.athlete, display_name=player.athlete.display_name, linescore=linescore
                )
                innings[linescore.period - 1].bowlers.append(bowl)



from pydantic import BaseModel, HttpUrl

from pycricinfo.models.source.api.common import CCBaseModel, Link, Position


class Flag(BaseModel):
    alt: str
    height: int
    href: HttpUrl
    rel: list[str]
    width: int


class Official(CCBaseModel):
    display_name: str
    flag: Flag
    links: Link
    order: int
    position: Position



from typing import Optional

from pydantic import Field

from pycricinfo.models.source.api.athelete import Athlete
from pycricinfo.models.source.api.common import Position, RefMixin


class Player(Athlete):
    date_of_birth: str
    date_of_death: Optional[str] = Field(default=None)
    active: bool
    gender: str
    position: Position
    country: int
    major_teams: list[RefMixin]
    debuts: list[RefMixin]



from datetime import datetime
from typing import Literal, Optional

from pydantic import AliasChoices, BaseModel, Field, computed_field

from pycricinfo.models.source.api.common import CCBaseModel, Link, MatchClass
from pycricinfo.models.source.api.league import League
from pycricinfo.models.source.api.linescores import TeamInningsDetails
from pycricinfo.models.source.api.match_note import MatchNote
from pycricinfo.models.source.api.official import Official
from pycricinfo.models.source.api.roster import TeamLineup
from pycricinfo.models.source.api.team import TeamWithColorAndLogos
from pycricinfo.models.source.api.venue import Venue


class MatchCompetitor(CCBaseModel):
    id: int
    winner: bool
    team: TeamWithColorAndLogos
    score: str = Field(
        description="One or two innings scores for the team, sometimes including the overs",
        examples=["421/5d", "150 & 130 (50.3 ov)"],
    )
    innings: list[TeamInningsDetails] = Field(validation_alias=AliasChoices("innings", "linescores"))
    home_or_away: Literal["home", "away"] = Field(validation_alias=AliasChoices("home_or_away", "homeAway"))


class MatchStatus(CCBaseModel):
    summary: str = Field(description="A summary of the result of the match", examples=["England won by 5 wickets"])


class MatchCompetiton(CCBaseModel):
    status: MatchStatus
    competitors: list[MatchCompetitor]
    limited_overs: bool
    reduced_overs: bool
    match_class: MatchClass = Field(validation_alias=AliasChoices("match_class", "class"))
    date: datetime = Field(description="The start date and time of the match in UTC", examples=["2024-07-26T10:00:00Z"])
    end_date: Optional[datetime] = Field(
        default=None, description="The end date and time of the match in UTC", examples=["2024-07-30T16:00:00Z"]
    )


class MatchHeader(CCBaseModel):
    id: int = Field(description="The Cricinfo ID for the match", examples=["1381212"])
    name: str = Field(description="The two teams competing in the match", examples=["West Indies v India"])
    short_name: str = Field(
        description="A short version of the two teams competing in the match", examples=["WI v IND"]
    )
    title: str = Field(
        description="A full title for the match, including teams, series, venue and dates",
        examples=["3rd Test, West Indies tour of England at Birmingham, Jul 26-28 2024"],
    )
    description: str = Field(description="Should match the 'title' field")
    competitions: list[MatchCompetiton] = Field(
        description="Details of the competition/match. There is always only 1 item."
    )
    links: list[Link] = Field(description="Links related to the match")
    leagues: list[League] = Field(description="The league(s) that this match is part of")

    @computed_field
    @property
    def competition(self) -> MatchCompetiton:
        """Source data has a list of competitions, but in reality there is only ever 1, so return it"""
        return self.competitions[0]

    def get_batting_innings_by_number(
        self, innings_number: int
    ) -> Optional[tuple[TeamWithColorAndLogos, TeamInningsDetails]]:
        """
        Get the batting innings data for a specific innings number.

        Parameters
        ----------
        innings_number : int
            The innings number to retrieve.

        Returns
        -------
        Optional[tuple[TeamWithColorAndLogos, TeamInningsDetails]]
            The team and innings details for the specified innings number, or None if not found.
        """
        for competitor in self.competition.competitors:
            for linescore in competitor.innings:
                if linescore.period == innings_number and linescore.is_batting:
                    return competitor.team, linescore


class MatchInfo(BaseModel):
    venue: Venue
    attendance: Optional[int] = None
    officials: list[Official]


class Match(CCBaseModel):
    notes: list[MatchNote]
    game_info: MatchInfo
    rosters: list[TeamLineup]
    header: MatchHeader

    @computed_field
    @property
    def teams(self) -> list[MatchCompetitor]:
        return self.header.competition.competitors

    @computed_field
    @property
    def start_date(self) -> datetime:
        return self.header.competition.date

    @computed_field
    @property
    def is_international(self) -> bool:
        return self.header.competition.match_class.international_class_id is not None

    @computed_field
    @property
    def summary(self) -> bool:
        """A summary of the result of the match, e.g.) 'England won by 5 wickets'"""
        return self.header.competitions[0].status.summary



from pydantic import Field

from pycricinfo.models.source.api.common import CCBaseModel, Link


class League(CCBaseModel):
    id: int = Field(description="The Cricinfo ID for the league", examples=["123"])
    name: str = Field(description="The name of the league", examples=["Indian Premier League"])
    short_name: str = Field(description="A short name for the league", examples=["IPL"])
    abbreviation: str = Field(description="The abbreviation for the league", examples=["England tour of India 2020-21"])
    league_type: str = Field(description="The type of league", examples=["Primary"])
    is_tournament: bool = Field(description="Whether the league is a tournament")
    links: list[Link] = Field(description="Links related to the league")



from typing import Annotated

from fastapi import Query


class PageAndInningsQueryParameters:
    def __init__(
        self,
        innings: Annotated[int, Query(description="Which innings of the game to get data from")] = 1,
        page: Annotated[int, Query(description="Which page of data to return")] = 1,
    ):
        self.page = page
        self.innings = innings



from typing import Literal

from pydantic import AliasChoices, BaseModel, Field, computed_field

from pycricinfo.models.source.api.athelete import Athlete
from pycricinfo.models.source.api.common import CCBaseModel, Position
from pycricinfo.models.source.api.linescores import PlayerInningsDetails
from pycricinfo.models.source.api.team import TeamWithColorAndLogos


class MatchPlayer(CCBaseModel):
    captain: bool
    active: bool
    active_name: str
    starter: bool
    athlete: Athlete
    position: Position
    linescores: list[PlayerInningsDetails]
    subbedIn: bool
    subbedOut: bool

    @computed_field
    @property
    def athlete_name(self) -> str:
        return self.athlete.display_name

    @computed_field
    @property
    def keeper(self) -> bool:
        return self.position.abbreviation == "WK"


class TeamLineup(BaseModel):
    home_or_away: Literal["home", "away"] = Field(validation_alias=AliasChoices("home_or_away", "homeAway"))
    winner: bool
    team: TeamWithColorAndLogos
    players: list[MatchPlayer] = Field(validation_alias=AliasChoices("players", "roster"))



from typing import Optional

from pycricinfo.models.source.api.common import CCBaseModel, Link, RefMixin


class Address(CCBaseModel):
    city: str
    state: Optional[str] = None
    zip_code: Optional[str] = None
    country: str
    summary: str


class Venue(CCBaseModel):
    id: str
    full_name: str
    short_name: str
    address: Address
    capacity: int
    grass: bool
    images: list[RefMixin]
    links: list[Link]



import re
from typing import Optional
from urllib.parse import quote

from bs4 import BeautifulSoup
from bs4._typing import _OneElement, _QueryResults

from pycricinfo.api_helper import get_request
from pycricinfo.config import BaseRoute, MatchTypeNames, get_settings
from pycricinfo.models.source.pages.series import MatchSeries, MatchTypeWithSeries


def get_match_types_in_season(
    season_name: str | int, type_filter: Optional[MatchTypeNames] = None
) -> list[MatchTypeWithSeries]:
    """
    Get the Cricinfo web page which lists all series in a given season, and parse out their details.

    Parameters
    ----------
    season_name : str | int
        The name of the season to get matches for, e.g. "2024" or "2020-21"

    Returns
    -------
    list[MatchTypeWithSeries]
        A list of match types, each containing a list of series in that match type for this season.
    """
    content = get_request(
        route=get_settings().page_routes.series_in_season,
        params={"season_name": quote(str(season_name))},
        base_route=BaseRoute.page,
        response_output_sub_folder="seasons",
    )

    match_types = parse_season_html(content)

    if type_filter:
        match_types = [m for m in match_types if m.name.lower() == type_filter.value.lower()]

    return match_types


def parse_season_html(content: str) -> list[MatchTypeWithSeries]:
    """
    Parse the content of the Cricinfo season page HTML file to extract series details.

    Parameters
    ----------
    content : str
        The season page content.

    Returns
    -------
    list[Series]
        A list of Series, with values for the title, id, link, and summary_url of a series in the season.
    """
    content = re.sub(r"^b\'|\'$", "", content)

    soup = BeautifulSoup(content, "html.parser")

    section_heads = soup.find_all("div", class_="match-section-head")

    match_types = []
    for section in section_heads:
        mt = _process_match_type_page_section(section)
        if mt:
            match_types.append(mt)

    return match_types


def _process_match_type_page_section(section: _OneElement) -> MatchTypeWithSeries | None:
    """
    Each page section representing a type of match should contian a h2 tag with the match type name,
    and then the following section will be a list of series of that type within the season

    Parameters
    ----------
    section : _OneElement
        The section of the series page for this match type

    Returns
    -------
    MatchType | None
        If the correct data was present, a parsed MatchType object, otherwise None
    """
    h2_tag = section.find("h2")
    if not h2_tag:
        return

    h2_text = h2_tag.text.strip()
    match_type = MatchTypeWithSeries(name=h2_text)  # TODO: Store enum name in object

    next_section = section.find_next_sibling("section", class_="series-summary-wrap")

    if next_section:
        series_blocks = next_section.find_all("section", class_="series-summary-block collapsed")

        series = _process_series_blocks(series_blocks)
        match_type.series = series
    return match_type


def _process_series_blocks(series_blocks: list[_QueryResults]) -> list[MatchSeries]:
    """
    Process all sections matched by BeautifulSoup which have the classes representing series within a match type.
    Iterate over the blocks and, within each, find the data for the series_id, title, and useful links.

    Parameters
    ----------
    series_blocks : list[_QueryResults]
        A list of blocks of data representing series in the match type

    Returns
    -------
    list[MatchSeries]
        A list of parsed series of matches
    """
    series_for_type = []
    for block in series_blocks:
        if "data-series-id" not in block.attrs:
            continue

        data_series_id = block["data-series-id"]

        series_link = block.find("a")
        if series_link:
            title = series_link.contents[0]
            title = re.sub(r"\\", "", title)
            title = re.sub(r"\s{2,}|\n|\r", " ", title).strip()

            data_series_id = block.get("data-series-id")
            summary_url = block.get("data-summary-url")

            link = series_link.get("href", "")

            series_id_regex_match = re.search(r"/series/[^/]+-(\d+)/", link)
            series_id = int(series_id_regex_match.group(1))

            s = MatchSeries(
                title=title, data_series_id=data_series_id, link=link, summary_url=summary_url, series_id=series_id
            )
            series_for_type.append(s)
    return series_for_type



from enum import Enum
from typing import Optional

from pydantic import Field, HttpUrl

from pycricinfo.models.source.api.common import CCBaseModel


class MatchNoteType(Enum):
    SERIES_NOTE = "seriesnote"
    POINTS = "points"
    MATCH_NUMBER = "matchnumber"
    SEASON = "season"
    MATCH_DAYS = "matchdays"
    TOSS = "toss"
    LIVE_COMMENTATOR = "livecommentator"
    LIVE_SCORER = "livescorer"
    MATCH_NOTE = "matchnote"
    CLOSE_OF_PLAY = "closeofplay"
    HOURS_OF_PLAY = "hoursofplay"
    PLAYER_REPLACEMENT = "playerreplacement"


class MatchNote(CCBaseModel):
    id: Optional[str | int] = Field(default=None)
    day_number: Optional[str | int] = Field(default=None)
    date: Optional[str] = Field(default=None)
    text: Optional[str] = Field(default=None)
    type: MatchNoteType
    href: Optional[HttpUrl] = Field(default=None)



from typing import Optional

from pydantic import BaseModel, computed_field, model_validator

from pycricinfo.models.source.api.common import CCBaseModel
from pycricinfo.models.source.api.dismissal import Dismissal


class BattingRecent(CCBaseModel):
    balls: int
    over_span: int
    runs: int


class PreferredShot(CCBaseModel):
    shot_name: str
    runs_summary: list[int]
    balls_faced: int
    runs: int


class BattingPvp(CCBaseModel):
    balls: int
    runs: int


class WagonZone(CCBaseModel):
    runs_summary: list[int]
    scoring_shots: int
    runs: int


class Wagon(BaseModel):
    long_leg: WagonZone
    backward_square_leg: WagonZone
    mid_wicket: WagonZone
    mid_on: WagonZone
    mid_off: WagonZone
    cover: WagonZone
    backward_point: WagonZone
    third: WagonZone


class BattingDetails(CCBaseModel):
    active: bool
    active_name: str
    order: int
    out_details: Dismissal
    pvp: BattingPvp
    runs_summary: list[int | str]  # 8 values: dots, singles, twos, threes, fours, X, sixes, X
    dot_ball_percentage: int
    batting_recent: BattingRecent
    preferred_shot: Optional[PreferredShot] = None
    scoring_shots: int
    control_percentage: int
    wagonZone: list[WagonZone]  # 8 zones: clockwise from long leg
    wagon: Wagon

    @computed_field
    @property
    def dismissal_text(self) -> Optional[str]:
        return self.out_details and self.out_details.short_text.replace("&dagger;", "\u271d").strip()

    @model_validator(mode="before")
    @classmethod
    def generate_wagon(cls, data: dict):
        wagon_fields = Wagon.model_fields.keys()

        wagon_zone = data.get("wagonZone", None)
        if wagon_zone:
            wagon = dict(zip(wagon_fields, wagon_zone))
            data["wagon"] = Wagon(**wagon)

        return data



from abc import ABC
from typing import Optional

from pydantic import AliasChoices, AliasGenerator, BaseModel, ConfigDict, Field, HttpUrl, model_validator
from pydantic.alias_generators import to_camel


class CCBaseModel(ABC, BaseModel):
    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel), validate_by_name=True, validate_by_alias=True
    )

    @model_validator(mode="before")
    @classmethod
    def set_empty_dicts_to_none(self, data: dict):
        if not data:
            return data

        for k, v in data.items():
            if isinstance(v, dict) and len(v) == 0:
                data[k] = None
        return data


class PagingModel(CCBaseModel):
    count: int
    page_index: int
    page_size: int
    page_count: int


class RefMixin(CCBaseModel):
    ref: Optional[HttpUrl] = Field(default=None, validation_alias=AliasChoices("ref", "$ref", "href"))


class DateMixin(CCBaseModel):
    date: Optional[str] = None


class DisplayNameMixin(BaseModel):
    display_name: str


class Link(CCBaseModel):
    language: Optional[str] = None
    rel: Optional[list[str] | str] = None
    href: HttpUrl
    text: str
    short_text: Optional[str] = None
    is_external: Optional[bool] = None
    is_premium: Optional[bool] = None


class Position(RefMixin):
    displayName: Optional[str] = None
    id: Optional[str] = None
    name: str
    abbreviation: Optional[str] = None


class Event(RefMixin, DateMixin): ...


class MatchClass(CCBaseModel):
    name: str
    event_type: str
    general_class_id: int
    general_class_card: Optional[str] = None
    international_class_id: Optional[int] = None
    international_class_card: Optional[str] = None



import uvicorn
from fastapi import FastAPI, Request
from fastapi.responses import JSONResponse
from fastapi_pagination import add_pagination

from pycricinfo.api.endpoints.match import router as match_router
from pycricinfo.api.endpoints.play_by_play import router as play_by_play_router
from pycricinfo.api.endpoints.player import router as player_router
from pycricinfo.api.endpoints.raw import router as raw_router
from pycricinfo.api.endpoints.scorecard import router as scorecard_router
from pycricinfo.api.endpoints.seasons import router as seasons_router
from pycricinfo.api.endpoints.team import router as team_router
from pycricinfo.config import get_settings
from pycricinfo.exceptions import CricinfoAPIException
from pycricinfo.utils import get_field_from_pyproject

app = FastAPI(
    version=get_field_from_pyproject("version"),
    title="pycricinfo API",
    swagger_ui_parameters={
        "defaultModelsExpandDepth": -1,
        "docExpansion": "none",
        "tryItOutEnabled": True,
    },
    description=get_field_from_pyproject("description"),
    openapi_tags=[
        {
            "name": "Cricinfo: API",
            "description": "Endpoints direct from the Cricinfo API, without any additional processing",
        },
    ],
)
add_pagination(app)

app.include_router(raw_router)
app.include_router(match_router)
app.include_router(play_by_play_router)
app.include_router(player_router)
app.include_router(seasons_router)
app.include_router(scorecard_router)
app.include_router(team_router)


@app.exception_handler(CricinfoAPIException)
def my_custom_exception_handler(request: Request, exc: CricinfoAPIException):
    return JSONResponse(
        status_code=exc.status_code,
        content=exc.output(),
    )


if __name__ == "__main__":
    uvicorn.run("pycricinfo.api.server:app", host="0.0.0.0", port=get_settings().port, reload=True)



from enum import Enum, auto
from functools import lru_cache

from pydantic import BaseModel
from pydantic_settings import BaseSettings


class MatchTypeNames(Enum):
    TESTS = "Tests"
    ODIs = "One-Day Internationals"
    T20Is = "Twenty20 Internationals"
    FIRST_CLASS = "First-class"
    LIST_A = "List A"
    T20 = "Twenty20"
    WOMENS_ODIs = "Women's One-Day Internationals"
    WOMENS_T20Is = "Women's Twenty20 Internationals"
    YOUTH_TESTS = "Youth Tests"
    YOUTH_ODIs = "Youth One-Day Internationals"
    WOMENS_T20 = "Women's Twenty20"
    TOUR = "Tour"
    MINOR_TOUR = "Minor tour"
    OTHER_MATCHES = "Other matches"
    OTHER_ODI_MATCHES = "Other one-day/limited-overs matches"
    OTHER_T20_MATCHES = "Other Twenty20 matches"


class BaseRoute(Enum):
    core: str = auto()
    site: str = auto()
    page: str = auto()


class CoreAPIRoutes(BaseModel):
    team: str = "events/0/teams/{team_id}"
    team_players: str = "cricket/teams/{team_id}/athletes"
    player: str = "teams/0/athletes/{player_id}"
    match_basic: str = "events/{match_id}"
    match_team: str = "leagues/0/events/{match_id}/competitions/{match_id}/competitors/{team_id}"
    match_summary: str = "{series_id}/summary?event={match_id}"
    league: str = "leagues/{league_id}"
    league_event: str = "leagues/{league_id}/events/{event_id}"
    play_by_play_page: str = "0/playbyplay?event={match_id}&page={page}&period={innings}"
    venue: str = "venues/{venue_id}"


class PageRoutes(BaseModel):
    series_in_season: str = "series/index.html?season={season_name};view=season"
    matches_in_series: str = "match/index/series.html?series={series_id}"


class PageHeaders(BaseModel):
    user_agent: str = "Mozilla/5.0 (Macintosh; Intel Mac OS X 10.15; rv:146.0) Gecko/20100101 Firefox/146.0"
    accept: str = "text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8"


class Settings(BaseSettings):
    core_base_route_v2: str = "http://core.espnuk.org/v2/sports/cricket/"
    site_base_route_v2: str = "http://site.api.espn.com/apis/site/v2/sports/cricket/"
    pages_base_route: str = "https://www.espncricinfo.com/ci/engine/"

    routes: CoreAPIRoutes = CoreAPIRoutes()
    page_routes: PageRoutes = PageRoutes()
    page_headers: PageHeaders = PageHeaders()
    api_response_output_folder: str = "responses"
    port: int = 8000

    match_classes: dict[int, str] = {
        1: "Test",
        2: "ODI",
        3: "T20I",
        4: "First Class",
        5: "List A",
        6: "T20",
        8: "Women's Test",
        9: "Women's ODI",
        10: "Women's T20I",
        11: "Combined Test, ODI and T20I",
        12: "Combined First Class, List A and T20",
        13: "All cricket, including minor cricket",
        20: "Under 19s Youth Test",
        21: "Under 19s Youth ODI",
        23: "Women's T20",
    }


@lru_cache
def get_settings():
    return Settings()



class CricinfoAPIException(Exception):
    status_code: int
    route: str
    content: dict | None = None

    def __init__(self, status_code: int, route: str, content: dict | None):
        self.status_code = status_code
        self.route = route
        self.content = content
        super().__init__()

    def output(self) -> dict:
        """
        Output the exception details as a dictionary.

        Returns
        -------
        dict
            A dictionary containing the route, and content.
        """
        return {
            "route": self.route,
            "content": self.content,
        }



import re

from bs4 import BeautifulSoup

from pycricinfo.config import BaseRoute, get_settings
from pycricinfo.api_helper import get_request
from pycricinfo.models.source.pages.series import MatchResult


def _clean_text(text: str) -> str:
    """
    Remove line breaks, extra spaces and normalise whitespace

    Parameters
    ----------
    text : str
        The text to clean

    Returns
    -------
    str
        The cleaned text
    """
    if not text:
        return ""
    text = text.replace("\\n", "")
    return re.sub(r"\s+", " ", text).strip()


def extract_match_ids_from_series(series_id: int | str) -> list[MatchResult]:
    """
    Extract match IDs from a series by fetching the series page and parsing the matches.

    Parameters
    ----------
    series_id : int | str
        The ID of the series to extract match IDs from.

    Returns
    -------
    list[MatchResult]
        A list of match results extracted from the series.
    """
    content = get_request(
        route=get_settings().page_routes.matches_in_series,
        params={"series_id": series_id},
        base_route=BaseRoute.page,
        response_output_sub_folder="series",
    )

    soup = BeautifulSoup(content, "html.parser")
    matches: list[MatchResult] = []

    match_blocks = soup.find_all("section", class_="default-match-block")

    for match in match_blocks:
        match_result_data = {}

        match_link = match.select_one(".match-no a")
        if match_link:
            href = match_link.get("href", "")
            scorecard_match = re.search(r"/scorecard/(\d+)/", href)
            if scorecard_match:
                match_result_data["id"] = scorecard_match.group(1)

            match_result_data["description"] = _clean_text(match_link.get_text())

        innings1 = match.find("div", class_="innings-info-1")
        if innings1:
            match_result_data["innings_1_info"] = _clean_text(innings1.get_text())

        innings2 = match.find("div", class_="innings-info-2")
        if innings2:
            match_result_data["innings_2_info"] = _clean_text(innings2.get_text())

        status = match.find("div", class_="match-status")
        if status:
            match_result_data["status"] = _clean_text(status.get_text())

        if "id" in match_result_data:
            matches.append(MatchResult(**match_result_data))

    return matches



from pydantic import AliasChoices, Field

from pycricinfo.models.source.api.common import CCBaseModel, MatchClass, RefMixin
from pycricinfo.models.source.api.venue import Venue


class MatchCompetitorBasic(CCBaseModel):
    id: int
    winner: bool
    home_or_away: str = Field(validation_alias=AliasChoices("home_or_away", "homeAway"))
    team: RefMixin
    score: RefMixin
    linescores: RefMixin
    roster: RefMixin
    leaders: RefMixin
    statistics: RefMixin
    record: RefMixin


class MatchCompetitionBasic(CCBaseModel):
    id: int
    description: str
    date: str
    end_date: str
    day_night: bool
    limited_overs: bool
    reduced_overs: bool
    match_class: MatchClass = Field(validation_alias=AliasChoices("match_class", "class"))
    venue: Venue
    competitors: list[MatchCompetitorBasic]


class MatchBasic(CCBaseModel):
    id: int
    name: str
    description: str
    short_name: str
    short_description: str
    season: RefMixin
    season_type: RefMixin
    venues: list[RefMixin]
    competitions: list[MatchCompetitionBasic]



from fastapi import APIRouter, Path, status

from pycricinfo.call_cricinfo_api import (
    get_scorecard,
)
from pycricinfo.models.output.scorecard import CricinfoScorecard

router = APIRouter(prefix="/scorecard", tags=["scorecard"])


@router.get(
    "/{series_id}/{match_id}",
    responses={status.HTTP_200_OK: {"description": "The match scorecard"}},
    summary="Get a match scorecard",
)
def scorecard(
    series_id: int = Path(description="The Series ID"), match_id: int = Path(description="The Match ID")
) -> CricinfoScorecard:
    return get_scorecard(series_id, match_id)



from abc import ABC
from typing import Literal, Optional

from pydantic import AliasChoices, BaseModel, Field, computed_field, field_validator

from pycricinfo.models.source.api.athelete import Athlete
from pycricinfo.models.source.api.common import CCBaseModel, RefMixin
from pycricinfo.models.source.api.statistics import PlayerStatisticsCategory, StatisticsCategory, TeamStatisticsCategory


class BaseInningsDetails(CCBaseModel, ABC):
    period: int
    statistics: StatisticsCategory

    def find(self, name: str) -> int | str | float:
        return self.statistics.find(name)


class PlayerInningsDetails(BaseInningsDetails):
    media_id: int = Field(validation_alias=AliasChoices("media_id", "mediaId"))
    statistics: PlayerStatisticsCategory

    @computed_field
    @property
    def batted(self) -> bool:
        return self.find("batted")

    @computed_field
    @property
    def bowled(self) -> bool:
        return self.find("bowled")


class PartnershipBatter(CCBaseModel):
    athlete: Athlete
    balls: str | int
    runs: str | int


class InningsState(BaseModel):
    overs: str | float
    runs: str | int
    wickets: str | int


class PartnershipFallOfWicketCommon(RefMixin, CCBaseModel, ABC):
    wicket_number: int
    fow_type: Literal["out", "end of innings", "not out", "retired not out"]
    runs: int

    @field_validator("fow_type", mode="before")
    @classmethod
    def strip_fow_type(cls, v):
        """
        In source data, fow_type has trailing whitespace on "retired not out", so trim it before validation.
        """
        if isinstance(v, str):
            return v.strip()
        return v


class Partnership(PartnershipFallOfWicketCommon):
    wicket_name: str
    overs: float
    run_rate: float
    start: InningsState
    end: InningsState
    batsmen: list[PartnershipBatter]


class FallOfWicket(PartnershipFallOfWicketCommon):
    wicket_over: float
    runs_scored: int
    balls_faced: int
    athlete: Athlete


class TeamInningsDetails(BaseInningsDetails):
    wickets: int
    runs: int
    overs: float
    is_batting: bool
    fours: Optional[int] = None
    sixes: Optional[int] = None
    description: str
    target: int
    follow_on: int
    statistics: Optional[TeamStatisticsCategory]
    partnerships: Optional[list[Partnership]] = None
    fall_of_wicket: Optional[list[FallOfWicket]] = Field(
        default=None, validation_alias=AliasChoices("fall_of_wicket", "fow")
    )

    @computed_field
    @property
    def declared(self) -> bool:
        return self.description.lower() == "declared"



# ruff: noqa F403

from .athelete import *
from .batting import *
from .bowling import *
from .commentary import *
from .common import *
from .dismissal import *
from .league import *
from .linescores import *
from .match_basic import *
from .match_note import *
from .match import *
from .official import *
from .player import *
from .roster import *
from .statistics import *
from .team import *
from .venue import *



import requests
from fastapi import APIRouter, Depends, Path, status

from pycricinfo.api.utils import PageAndInningsQueryParameters
from pycricinfo.api_helper import get_request
from pycricinfo.config import BaseRoute, get_settings

router = APIRouter(prefix="/raw", tags=["Cricinfo: API"])


@router.get(
    "/match/{match_id}",
    responses={status.HTTP_200_OK: {"description": "The basic match data"}},
    summary="Get basic match data from the '/events' API",
)
def match_basic(match_id: int = Path(description="The Match ID")):
    return get_request(get_settings().routes.match_basic, params={"match_id": match_id})


@router.get(
    "/match/{match_id}/team/{team_id}",
    responses={status.HTTP_200_OK: {"description": "The basic match data"}},
    summary="Get a match's Team",
)
def match_team(match_id: int = Path(description="The Match ID"), team_id: int = Path(description="The Team ID")):
    response = requests.get(
        f"http://core.espnuk.org/v2/sports/cricket/leagues/0/events/{match_id}/competitions/{match_id}/competitors/{team_id}"
    ).json()
    return response


@router.get(
    "/match/{match_id}/team/{team_id}/roster",
    responses={status.HTTP_200_OK: {"description": "The basic match data"}},
    summary="Get a match Team's roster",
)
def match_team_roster(match_id: int = Path(description="The Match ID"), team_id: int = Path(description="The Team ID")):
    response = requests.get(
        f"http://core.espnuk.org/v2/sports/cricket/leagues/0/events/{match_id}/competitions/{match_id}/competitors/{team_id}/roster"
    ).json()
    return response


@router.get(
    "/match/{match_id}/team/{team_id}/innings",
    responses={status.HTTP_200_OK: {"description": "The basic match data"}},
    summary="Get a match Team's innings",
)
def match_team_all_innings(
    match_id: int = Path(description="The Match ID"),
    team_id: int = Path(description="The Team ID"),
):
    response = requests.get(
        f"http://core.espnuk.org/v2/sports/cricket/leagues/0/events/{match_id}/competitions/{match_id}/competitors/{team_id}/linescores"
    ).json()
    return response


@router.get(
    "/match/{match_id}/team/{team_id}/innings/{innings}",
    responses={status.HTTP_200_OK: {"description": "The basic match data"}},
    summary="Get a match Team's innings",
)
def match_team_innings(
    match_id: int = Path(description="The Match ID"),
    team_id: int = Path(description="The Team ID"),
    innings: int = Path(description="The innings number"),
):
    response = requests.get(
        f"http://core.espnuk.org/v2/sports/cricket/leagues/0/events/{match_id}/competitions/{match_id}/competitors/{team_id}/linescores/0/{innings}"
    ).json()
    return response


@router.get(
    "/match/{match_id}/team/{team_id}/player/{player_id}/innings",
    responses={status.HTTP_200_OK: {"description": "The basic match data"}},
    summary="Get a match Team's innings",
)
def match_player_all_innings(
    match_id: int = Path(description="The Match ID"),
    team_id: int = Path(description="The Team ID"),
    player_id: int = Path(description="The Player ID"),
):
    response = requests.get(
        f"http://core.espnuk.org/v2/sports/cricket/leagues/0/events/{match_id}/competitions/{match_id}/competitors/{team_id}/roster/{player_id}/linescores"
    ).json()
    return response


@router.get(
    "/match/{match_id}/play_by_play_page",
    responses={status.HTTP_200_OK: {"description": "The match summary"}},
    summary="Get a page of ball-by-ball data",
)
def match_play_by_play(match_id: int = Path(description="The Match ID"), pi: PageAndInningsQueryParameters = Depends()):
    return get_request(
        get_settings().routes.play_by_play_page,
        {"match_id": match_id, "page": pi.page, "innings": pi.innings},
        BaseRoute.site,
    )


@router.get(
    "/match_summary/{match_id}",
    responses={status.HTTP_200_OK: {"description": "The match summary"}},
    summary="Get a match summary",
)
def match_summary(match_id: int = Path(description="The Match ID")):
    return get_request(get_settings().routes.match_summary, params={"match_id": match_id}, base_route=BaseRoute.site)


@router.get(
    "/series/{series_id}/match/{match_id}/team/{team_id}/statistics",
    responses={status.HTTP_200_OK: {"description": "The basic match data"}},
    summary="Get a match Team's statistics",
)
def match_team_statistics(
    series_id: int = Path(description="The Series ID"),
    match_id: int = Path(description="The Match ID"),
    team_id: int = Path(description="The Team ID"),
):
    response = requests.get(
        f"http://core.espnuk.org/v2/sports/cricket/leagues/{series_id}/events/{match_id}/competitions/{match_id}/competitors/{team_id}/statistics"
    ).json()
    return response


@router.get(
    "/series/{series_id}/match/{match_id}/team/{team_id}/player/{player_id}/innings/{innings}/statistics",
    responses={status.HTTP_200_OK: {"description": "The basic match data"}},
    summary="Get statistics for a player's innings",
)
def match_player_innings(
    series_id: int = Path(description="The Series ID"),
    match_id: int = Path(description="The Match ID"),
    team_id: int = Path(description="The Team ID"),
    player_id: int = Path(description="The Player ID"),
    innings: int = Path(description="The innings number"),
):
    response = requests.get(
        f"http://core.espnuk.org/v2/sports/cricket/leagues/{series_id}/events/{match_id}/competitions/{match_id}/competitors/{team_id}/roster/{player_id}/linescores/0/{innings}/statistics/0"
    ).json()
    return response


@router.get(
    "/venue/{venue_id}",
    responses={status.HTTP_200_OK: {"description": "A Venue's data"}},
    summary="Get a Venue",
)
def venue(venue_id: int = Path(description="The Venue ID")):
    return get_request(get_settings().routes.venue, params={"venue_id": venue_id})


@router.get(
    "/league/{league_id}",
    responses={status.HTTP_200_OK: {"description": "A League's data"}},
    summary="Get a League",
)
def league(league_id: int = Path(description="The League ID")):
    return get_request(get_settings().routes.league, params={"league_id": league_id})


@router.get(
    "/league/{league_id}/event/{event_id}",
    responses={status.HTTP_200_OK: {"description": "A League's data"}},
    summary="Get an event (aka: a series of matches) in a League",
)
def series_in_league(
    league_id: int = Path(description="The League ID"), event_id: int = Path(description="The Event ID")
):
    return get_request(get_settings().routes.league_event, params={"league_id": league_id, "event_id": event_id})


@router.get(
    "/team/{team_id}", responses={status.HTTP_200_OK: {"description": "The Team data"}}, summary="Get Team data"
)
def team(team_id: int = Path(description="The Team ID")):
    return get_request(get_settings().routes.team, params={"team_id": team_id})


@router.get("/player/{player_id}", responses={status.HTTP_200_OK: {"description": "The Player"}}, summary="Get Player")
def player(player_id: int = Path(description="The Player ID")):
    return get_request(get_settings().routes.player, params={"player_id": player_id})



from fastapi import APIRouter, Path, status

from pycricinfo.call_cricinfo_api import get_player
from pycricinfo.models.source.api.player import Player

router = APIRouter(prefix="/player", tags=["player"])


@router.get("/{player_id}", responses={status.HTTP_200_OK: {"description": "The Player"}}, summary="Get Player")
def player(player_id: int = Path(description="The Player ID")) -> Player:
    return get_player(player_id)



from abc import ABC
from typing import Optional

from pydantic import BaseModel, Field, HttpUrl

from pycricinfo.models.source.api.common import CCBaseModel, DisplayNameMixin, Link, RefMixin


class Style(CCBaseModel):
    description: str
    short_description: str
    type: str


class Headshot(BaseModel):
    href: HttpUrl
    rel: list[str]


class ShortNameMixin(BaseModel):
    short_name: str


class FullNameMixin(BaseModel):
    full_name: Optional[str] = Field(default=None)


class FirstNameMixin(BaseModel):
    first_name: Optional[str] = Field(default=None)


class LastNameMixin(BaseModel):
    last_name: str


class AthleteCommon(RefMixin, FullNameMixin, DisplayNameMixin, ABC):
    id: str | int = Field(description="The unique identifier for the athlete")


class AthleteWithNameAndShortName(AthleteCommon, ShortNameMixin):
    name: str = Field(description="The full name of the athlete")


class AthleteWithFirstAndLastName(AthleteCommon, FirstNameMixin, LastNameMixin): ...


class Athlete(AthleteWithFirstAndLastName):
    guid: Optional[str] = None
    uid: str
    name: str
    style: Optional[list[Style]] = None
    batting_name: str
    fielding_name: str
    headshot: Optional[Headshot] = None
    links: list[Link]

    def __str__(self) -> str:
        return f"{self.name} ({self.uid})"



from typing import Optional

from pydantic import AliasChoices, BaseModel, Field, computed_field

from pycricinfo.models.source.api.athelete import AthleteWithNameAndShortName as Athlete
from pycricinfo.models.source.api.common import CCBaseModel, PagingModel
from pycricinfo.models.source.api.team import TeamWithName as Team


class PlayType(BaseModel):
    id: str
    description: str


class Bowler(BaseModel):
    athlete: Optional[Athlete] = Field(default=None)
    team: Team
    maidens: int
    balls: int
    wickets: int
    overs: float
    conceded: int


class Batsman(CCBaseModel):
    athlete: Athlete
    team: Team
    total_runs: int
    faced: int
    fours: int
    runs: int
    sixes: int


class Innings(CCBaseModel):
    id: str
    run_rate: float
    remaining_balls: int
    byes: int
    number: int
    balls: int
    no_balls: int
    wickets: int
    leg_byes: int
    ball_limit: int
    target: int
    session: int
    day: int
    fall_of_wickets: int
    trailBy: int
    leadBy: int
    remaining_overs: float
    total_runs: int
    wides: int
    runs: int

    @computed_field
    @property
    def score(self) -> str:
        return f"{self.runs}/{self.wickets}"


class Over(CCBaseModel):
    ball: int
    balls: int
    complete: bool
    limit: float
    maiden: int
    no_ball: int
    wide: int
    leg_byes: int
    byes: int
    number: int
    runs: int
    wickets: int
    overs: float
    actual: float
    unique: float


class BowlerAthlete(BaseModel):
    athlete: Athlete


class BatsmanAthlete(BaseModel):
    athlete: Athlete


class Dismissal(CCBaseModel):
    dismissal: bool
    bowled: bool
    type: str
    bowler: BowlerAthlete
    batsman: BatsmanAthlete
    text: str
    minutes: Optional[int | str] = None  # TODO: Can be empty string - parse to null in that case
    retired_text: str


class CommentaryItem(CCBaseModel):
    id: str
    clock: str
    date: str
    play_type: PlayType
    team: Team
    media_id: int
    period: int
    periodText: str
    preText: str
    text: str
    post_text: str
    short_text: str
    home_score: str
    away_score: str
    score_value: int
    sequence: int
    athletes_involved: list[Athlete]
    bowler: Bowler
    other_bowler: Optional[Bowler] = None
    batsman: Batsman
    other_batsman: Batsman
    current_innings_score: Innings = Field(validation_alias=AliasChoices("current_innings_score", "innings"))
    over: Over
    dismissal: Dismissal
    bbb_timestamp: int

    @computed_field
    @property
    def summary(self) -> str:
        return f"{self.over.overs}: {self.short_text} - {self.current_innings_score.score}"


class Commentary(PagingModel):
    deliveries: list[CommentaryItem] = Field(validation_alias=AliasChoices("items", "deliveries"))


class APIResponseCommentary(BaseModel):
    commentary: Commentary



from fastapi import APIRouter, Path, status

from pycricinfo.call_cricinfo_api import get_team
from pycricinfo.models.source.api.team import TeamFull

router = APIRouter(prefix="/team", tags=["team"])


@router.get("/{team_id}", responses={status.HTTP_200_OK: {"description": "The Team data"}}, summary="Get Team data")
def team(team_id: int = Path(description="The Team ID")) -> TeamFull:
    return get_team(team_id)



from fastapi import APIRouter, Path, Query, status

from pycricinfo.config import MatchTypeNames
from pycricinfo.models.source.pages.series import MatchResult, MatchTypeWithSeries
from pycricinfo.search.seasons import get_match_types_in_season
from pycricinfo.search.series import extract_match_ids_from_series

router = APIRouter(prefix="", tags=["seasons"])


@router.get(
    "/season/{season_name}",
    responses={
        status.HTTP_200_OK: {
            "description": "A list of match types, each containing a list of series in that match type for this season"
        }
    },
    summary="Get a list of match types, each containing a list of series in that match type for this season",
)
def match_types_in_season(
    season_name: int | str = Path(description='The name of the season to get matches for, e.g. "2024" or "2020-21"'),
    match_type_name: MatchTypeNames = Query(
        default=None, description="Filter the response to just matches of the named type"
    ),
) -> list[MatchTypeWithSeries]:
    season_name = season_name.replace("-", "/") if isinstance(season_name, str) else season_name
    match_types = get_match_types_in_season(season_name, match_type_name)

    return match_types


@router.get(
    "/series/{data_series_id}",
    responses={status.HTTP_200_OK: {"description": "A list of IDs of the matches in the supplied series"}},
    summary="Get a list of IDs of the matches in the supplied series",
)
def match_ids_in_series(data_series_id: int = Path(description="The ID of a series")) -> list[MatchResult]:
    return extract_match_ids_from_series(data_series_id)
